import pymysql,clickhouse_driver,pymssql,threading,datetime,time,traceback

class DatabaseConnectBean(object):
    '''数据库连接对象'''
    def __init__(self,connect, checkConnectValidFunc):
        self.connect = connect
        self.lastTime = datetime.datetime.now()
        self.isUse = False
        self.checkConnectValidFunc = checkConnectValidFunc

    def isValid(self):
        if self.connect is None:
            return False
        try:
           return self.checkConnectValidFunc(self.connect)
        except Exception as e:
            traceback.print_exc()
            self.closeConnect()
            return False

    def getConnect(self):
        self.isUse = True
        self.lastTime = datetime.datetime.now()
        return self.connect

    def putConnect(self):
        self.isUse = False
        self.lastTime = datetime.datetime.now()

    def checkOutTime(self, timeout):
        if self.connect is None :
            return True
        if self.isUse is False and (datetime.datetime.now() - self.lastTime).seconds > timeout:
            return True
        return False

    def closeConnect(self):
        try:
            # print("关闭数据库连接",self,self.connect)
            if self.connect is not None:
                self.connect.close()
        except Exception as e:
            traceback.print_exc()
        finally:
            self.connect = None

class DatabaseClearThread(object):
    '''数据库连接池清理对象'''
    def __init__(self,idleTime=5*60):
        self.idleTime = idleTime
        self.databaseManagerList = []
        threading.Thread(name='数据库连接池空闲清理', target=self.run, daemon=True).start()

    def run(self):
        while True:
            time.sleep(self.idleTime)
            for dbmng in self.databaseManagerList:
                # print("连接检测 %s>> %s" % (threading.current_thread().name,str(dbmng.flag)))
                dbmng.clearInvalidConnect()

class DatabaseManagerBean(object):
    '''数据库管理对象'''
    def __init__(self,databaseClearThread,databaseType,host,port,username,password,databaseName,charset='uft8'):
        if databaseType ==  'mysql' or databaseType == 'clickHouse' or databaseType == 'sqlServer':
            self.flag = "数据库(%s,%s,%s,%s,%s,%s)-%s" \
                        % (databaseType, host, port,username, password,databaseName, datetime.datetime.now().strftime('%Y%m%d%H%M%S'))
        else:
            raise RuntimeError("不支持的数据库类型: %s" % databaseType)

        self.databaseType = databaseType
        self.host = host
        self.port = int(port)
        self.username = username
        self.password = password
        self.databaseName = databaseName
        self.charset = charset
        self.idleTime = databaseClearThread.idleTime
        # Map<线程,数据库连接对象>
        self._thread_databaseConnect_map = {}
        databaseClearThread.databaseManagerList.append(self)

    def clearInvalidConnect(self):
        '''清理无效连接'''
        for threadUse in list(self._thread_databaseConnect_map.keys()):
            try:
                connectHolder = self._thread_databaseConnect_map[threadUse]
                if connectHolder.checkOutTime(self.idleTime):
                    connectHolder.closeConnect()
                    self._thread_databaseConnect_map.pop(threadUse)
            except Exception as e:
                traceback.print_exc()

    def _databaseConnectCreate(self):
        '''创建数据库连接'''
        if self.databaseType == 'mysql':
            return pymysql.connect(
                host=self.host,
                port=self.port,
                user=self.username,
                password=self.password,
                db=self.databaseName,
                autocommit=False,
                read_timeout=self.idleTime,
                write_timeout=self.idleTime,
                charset=self.charset)

        if self.databaseType == "clickHouse":
            return clickhouse_driver.connect(
                host=self.host,
                port=self.port,
                user=self.username,
                password=self.password,
                database=self.databaseName,
                connect_timeout=self.idleTime,
                compression=False)

        if self.databaseType == 'sqlServer':
            return  pymssql.connect(
                host=self.host,
                user=self.username,
                password=self.password,
                database=self.databaseName,
                timeout=self.idleTime,
                login_timeout=self.idleTime,
                charset=self.charset)
        return None

    def _getConnectCheckFunc(self):
        '''获取连接是否正常的检测函数'''
        func = None
        if self.databaseType == 'mysql':
            def _executeFunc(connect):
                try:
                    connect.ping(reconnect=True)
                    return True
                except:
                    return False
            func = _executeFunc
        if self.databaseType == 'clickHouse':
            def _executeFunc(connect):
                try:
                    return True if connect.is_closed is False else False
                except:
                    return False
            func = _executeFunc
        if self.databaseType == 'sqlServer':
            def _executeFunc(connect):
                try:
                    return connect._conn.connected
                except Exception as e:
                    return False
            func = _executeFunc
        return func

    def threadGetDatabaseConnect(self):
        '''线程获取数据库连接'''
        try:
            connectHolder = self._thread_databaseConnect_map.get(threading.currentThread())
            if connectHolder is None or connectHolder.isValid() is False:
                db_connect = self._databaseConnectCreate()
                connectHolder = DatabaseConnectBean(db_connect, self._getConnectCheckFunc())
                self._thread_databaseConnect_map[threading.currentThread()] = connectHolder
            if connectHolder is None:raise Exception('%s 无法获取数据库连接,当前线程:%s' % (self.flag, threading.currentThread()))

        except Exception as e:
            traceback.print_exc()
            return None, None
        return connectHolder, connectHolder.getConnect()

    def dataErrorCallback(self, e, dataErrorCallback):
        if dataErrorCallback is not None:
            try:
                from pymysql.err import DataError
                if type(e) == DataError:
                    dataErrorCallback(e.args[1])
            except:
                pass

    def query(self, sql, isPrint = False, descriptionList = None):
        '''查询'''
        connectHolder, connect = self.threadGetDatabaseConnect()
        if connect is None: return []
        cursor = None
        try:
            import datetime
            t = datetime.datetime.now()
            cursor = connect.cursor()
            cursor.execute(sql)
            datas = cursor.fetchall()
            connect.commit()  # 解决查询存在缓存问题
            t1 = datetime.datetime.now()
            diff = t1 - t
            if isPrint:
                print('%s\n查询SQL: %s\n数据大小: %d\n时长: %s' % (self.flag, sql, len(datas), str(diff)))
            if descriptionList is not None:
                for des in cursor.description:
                    descriptionList.append(des[0])
            return datas
        except Exception as e:
            traceback.print_exc()
            print('%s 查询错误\nSQL: %s\nERROR: %s' % (self.flag, sql,e ))
        finally:
            try:
                if cursor is not None:
                    cursor.close()
            except:
                pass
            connectHolder.putConnect()
        return []

    def execute(self,sql,dataErrorCallback=None):
        '''执行SQL'''
        connectHolder, connect = self.threadGetDatabaseConnect()
        if connect is None: return -1

        cursor = None
        try:
            cursor = connect.cursor()
            cursor.execute(sql)
            connect.commit()
            return cursor.rowcount
        except Exception as e:
            traceback.print_exc()
            try:
                if connect is not None: connect.rollback()
            except: pass
            print('%s 执行错误\nSQL: %s\nERROR: %s' % (self.flag, sql, e))

            self.dataErrorCallback(e,dataErrorCallback)
        finally:
            try:
                if cursor is not None: cursor.close()
            except: pass
            connectHolder.putConnect()
        return -1

    def execute_transaction(self, sqlList, dataErrorCallback=None):
            '''事务执行SQL'''
            if len(sqlList) == 0 : return -1

            connectHolder, connect = self.threadGetDatabaseConnect()
            if connect is None: return -1

            cursor = None
            try:
                rows = 0
                cursor = connect.cursor()
                for index in range(0,len(sqlList)):
                    sql = sqlList[index]
                    try:
                        cursor.execute(sql)
                        rows += cursor.rowcount
                    except Exception as  e:
                        print('%s 事务执行错误\nINDEX: %d, SQL: %s\nERROR: %s' % (self.flag, index, sql, e))
                        raise e

                connect.commit()
                return rows
            except Exception as e:
                traceback.print_exc()
                try:
                    if connect is not None: connect.rollback()
                except:
                    pass
                self.dataErrorCallback(e, dataErrorCallback)
            finally:
                try:
                    if cursor is not None: cursor.close()
                except:
                    pass
                connectHolder.putConnect()
            return -1

    def execute_batch(self, sql, paramTupleList, dataErrorCallback=None):
        '''执行SQL'''
        if len(paramTupleList) == 0: return -1
        connectHolder, connect = self.threadGetDatabaseConnect()
        if connect is None: return -1
        cursor = None
        try:
            cursor = connect.cursor()
            cursor.executemany(sql, paramTupleList)
            connect.commit()
            return cursor.rowcount
        except Exception as e:
            import traceback
            traceback.print_exc()
            try:
                if connect is not None: connect.rollback()
            except:pass
            print('%s 批量执行错误\nSQL: %s\n批量条数:%d\nERROR: %s' % (self.flag, sql, len(paramTupleList), e))

            self.dataErrorCallback(e,dataErrorCallback)
        finally:
            try:
                if cursor is not None: cursor.close()
            except: pass
            connectHolder.putConnect()
        return -1

    def __str__(self) -> str:
        return super().__str__() + self.flag








'''
数据库工具类
'''

import os
import traceback
import threading
import datetime
from lsptoolbox.common_utils import strToDate,strToDatetime,decimalRound



pid_index = '%02d' % (os.getpid() % 99)
mLock = threading.RLock()

# 获取当前时间戳 毫秒
def _getCurrentTimestamp():
    import time
    timestamp = int(round(time.time() * 1000))
    return timestamp

def _getCurrentTimestampSec():
    import time
    timestamp = int(time.time())
    return timestamp

# 每秒产生最多9999个ID
lastTime = _getCurrentTimestamp()
seqIndex = 1
# 时间戳生成唯一ID
def genTimestampDataID():
    global mLock,lastTime,seqIndex
    mLock.acquire()
    try:
        curTime = _getCurrentTimestamp()
        if curTime - lastTime < 1000 and seqIndex <= 9999:
            num_str = '%04d' % seqIndex
            seqIndex = seqIndex + 1
            return '%s%s%s' % (lastTime, num_str, pid_index)
        else:
            lastTime = curTime
            seqIndex = 1
            return genTimestampDataID()

    except Exception as e:
        traceback.print_exc()
        print('获取时间戳产生的数据标识错误',e)
    finally:
        mLock.release()

# 每秒产生最多999个ID且可设置后缀
lastTimeSec = _getCurrentTimestampSec()
seqIndexSec = 1

def genTimestampDataIDSec(suffix='0000%s'% pid_index ):
    '''时间戳生成唯一ID'''
    global mLock, lastTimeSec, seqIndexSec
    mLock.acquire()
    try:
        curTimeSec = _getCurrentTimestampSec()
        if curTimeSec == lastTimeSec and seqIndexSec <= 999:
            num_str = '%03d' % seqIndexSec
            seqIndexSec = seqIndexSec + 1
            return '%s%s%s' % (lastTimeSec, num_str, suffix)
        else:
            while True:
                if _getCurrentTimestampSec() > curTimeSec: break
            lastTimeSec = _getCurrentTimestampSec()
            seqIndexSec = 1
            return genTimestampDataIDSec(suffix)
    except Exception as e:
        traceback.print_exc()
        print('获取时间戳产生的数据标识错误', e)
    finally:
        mLock.release()

# 数据行转map
def rowConvertDict(rows,*dictKeys):
    '''按照顺序对 指定键 赋值'''
    dict = {}
    if type(dictKeys[0]) == list:
        dictKeys = dictKeys[0]
    try:
        for i in range(len(dictKeys)):
            try:
                keyObj = dictKeys[i]
                if type(keyObj) == str:
                    _v = rows[i]
                    if isinstance(_v,datetime.datetime):
                        _v = datetime.datetime.strftime(_v, '%Y-%m-%d %H:%M:%S')
                    dict[keyObj] = str(_v if _v is not None else '')
                if type(keyObj) == tuple:
                    kstr = keyObj[0]
                    index = keyObj[1]
                    if len(keyObj) >= 3:
                        func = keyObj[2]
                    else:
                        func = lambda row_index_v: row_index_v
                    _v = func(rows[index])
                    dict[kstr] = str(_v if _v is not None else '')
            except :
                traceback.print_exc()
    except:
        traceback.print_exc()
    return dict

def rowConvertDict_str(rows,dictKeys_str):
    return rowConvertDict(rows,dictKeys_str.split(","))

def linesConvertDict(lines,descriptionList):
    list = []
    for rows in lines:
        maps = rowConvertDict(rows, descriptionList)
        list.append(maps)
    return list

def MysqlRowDateToJsonEncoderFunc(rowsDict):
    from lsptoolbox.common_utils import objToJson
    return objToJson(rowsDict)

def pauseCreateTableDDL(createDDL, isPrint=False):
    try:
        from ddlparse import DdlParse
        createDDL = createDDL.replace('Nullable(DateTime)', 'DateTime')
        createDDL = createDDL.replace('Array(String)', 'String')
        table = DdlParse().parse(createDDL, source_database=DdlParse.DATABASE.mysql)
        fieldTupleList = []
        for col in table.columns.values():
            if isPrint : print(table.name, col.name, col.data_type, col.comment)
            fieldTupleList.append((col.comment, col.name, col.data_type))
        return table.name , fieldTupleList
    except: return None,None


CLICKHOUSE_TYPE_MAPS = {
    'string':(str,''),
    'uint8': (int,0),
    'uint16':(int,0),
    'uint32':(int,0),
    'uint64':(int,0),
    'uint128':(int,0),
    'uint256':(int,0),
    'int8':  (int,0),
    'int16': (int,0),
    'int32': (int,0),
    'int64': (int,0),
    'int128': (int,0),
    'int256': (int,0),
    'date': (strToDate, ''),
    'datetime': (strToDatetime, ''),
    'date64': (strToDate, ''),
    'datetime64': (strToDatetime, ''),
    'float32': (float,0.0),
    'float64': (float,0.0),
    'decimal': (decimalRound,'0.0000'),
    'decimal32': (decimalRound,'0.0000'),
    'decimal64': (decimalRound,'0.0000'),
    'decimal128': (decimalRound,'0.0000')
}

def generateClickHouseInsertDDL(tableName, fieldTupleList):
    '''按表生成插入语句'''
    insertFieldKeyAndTypeTupleList = []
    insertFieldNames = []
    for tup in fieldTupleList:
        if str(tup[0]).startswith('#'): continue
        insertFieldNames.append(tup[1])
        tup_type_defv = CLICKHOUSE_TYPE_MAPS.get(str(tup[2]).lower(), None)
        insertFieldKeyAndTypeTupleList.append((tup[0], tup_type_defv[0], tup_type_defv[1]))
    insertSQL = '''INSERT INTO %s ( %s ) VALUES''' % (tableName, ','.join(insertFieldNames))
    return insertSQL , insertFieldKeyAndTypeTupleList






def cmdLineInput(tips='请输入应用启动类型: '):
    import re
    str = input(tips).lower()
    arrays = re.split(r"[ ]+", str)
    if arrays[len(arrays) - 1] == '':
        del arrays[len(arrays) - 1]
    if len(arrays) == 0:
        return None,(),()
    appType = arrays[0]
    params = []
    for index in range(1, len(arrays)):
        params.append(arrays[index])
    params = []
    for index in range(1, len(arrays)):
        params.append(arrays[index])
    return appType,tuple(params),tuple(arrays)

def getNowTimeStr(timeFormat='%Y-%m-%d %H:%M:%S'):
    '''获取当前时间字符串'''
    import datetime
    return datetime.datetime.strftime(datetime.datetime.now(), timeFormat)

def printFormat(format_str, *args):
    if len(args) > 0:
        print(format_str % args)
    else:
        print(format_str)

def isPortInUse(port, ip='127.0.0.1'):
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((ip, port))
        s.shutdown(2)
        return True
    except:
        return False

def validConfigFile(config, func):
    '''效验配置文件'''
    sections = config.sections()
    for section in sections:
        options = config.options(section)
        for option in options:
            value = config.get(section, option)
            new_value = func(value)
            if new_value != value:
                config.set(section, option, new_value)

def listDuplicate(data_list):
    '''列表去重'''
    from functools import reduce
    run_function = lambda x, y: x if y in x else x + [y]
    return reduce(run_function, [[], ] + data_list)

def convertFloatNum(_v):
    if type(_v) == str and len(_v) == 0:
        return 0
    else:
        return float(_v)

def createXLSX(filePath,titleArr,sheetName='Sheet1'):
    try:
        import xlsxwriter
        workbook = xlsxwriter.Workbook(filePath)  # 创建一个Excel文件
        worksheet = workbook.add_worksheet(sheetName)  # 创建一个sheet
        worksheet.write_row(0, 0, titleArr)  # title 写入Excel
        workbook.close()
    except Exception as e:
        print("EXCEL创建失败", e)

def writeXLSX(filePath,rowsData,nextRow=-1,sheetName='Sheet1'):
    try:
        import openpyxl
        workbook = openpyxl.load_workbook(filePath)
        worksheet = workbook.get_sheet_by_name(sheetName)
        if nextRow == -1:
            nextRow = worksheet.max_row + 1  # 获得行数
        rowIndex = 1
        for it in rowsData:
            worksheet.cell(nextRow, rowIndex).value = it
            rowIndex = rowIndex + 1
        workbook.save(filePath)
    except Exception as e:
        print("EXCEL写入失败",e)

def writeXLSXBatch(filePath,rowsDataList,sheetName='Sheet1'):
    try:
        import openpyxl
        workbook = openpyxl.load_workbook(filePath)
        worksheet = workbook.get_sheet_by_name(sheetName)
        startRow = worksheet.max_row + 1  # 获得行数
        for rowsData in rowsDataList:
            rowIndex = 1
            for it in rowsData:
                worksheet.cell(startRow, rowIndex).value = it
                rowIndex = rowIndex + 1
            print("当前写入行(%d),总行数(%d)" % (startRow, len(rowsDataList)))
            startRow=startRow+1
        workbook.save(filePath)
    except Exception as e:
        print("EXCEL批量写入失败",e)

def excelConvertTypeValue(valueType, value):
    # print('excelConvertTypeValue', valueType, type(value), value)
    try:
        # 单元格的ctype属性为0时，对应的python格式为空字符串：''
        if valueType == 0:
            return ''
        # 单元格的ctype属性为1时，对应的python格式为字符串
        elif valueType == 1:
            return value.strip()
        # 单元格的ctype属性为2时，对应的python格式为float和int
        elif valueType == 2:
            if value % 1 == 0.0:
                return str(int(value))
            else:
                return str(value)
        # 单元格的ctype属性为3时，对应的python格式为datetime
        elif valueType == 3:
            import datetime,xlrd
            dateStr = "%.4d-%.2d-%.2d %.2u:%.2u:%.2u" % xlrd.xldate_as_tuple(value, 0)
            return dateStr
        # 单元格的ctype属性为4时，对应的python格式为Bool
        elif valueType == 4:
            return True if value == 1 else False
    except Exception as e:
        # print('excel转换单元格数值错误', valueType, type(value), value, e)
        return ''
    return value

def excelSheetRead(tableName,table,contentList,filterContentList,dict_convert_head = None,filter_rows = None,filter_rows_args=None):
    if table.nrows == 0: return
    # 表头 第一行
    headsRows = table.row_values(0)
    for i in range(1, table.nrows):
        excel_rows = table.row_values(i)
        excel_rows_type = table.row_types(i)
        # 每行数据
        rows_dict = {"excel_line": '%s : %d' % (tableName,i)}
        # 读取行内容
        for j in range(0, len(headsRows)):
            key = headsRows[j]
            if len(str(key)) == 0: continue
            value = excel_rows[j]
            value_type = excel_rows_type[j]
            # 转换excel字段
            rows_dict[key] = excelConvertTypeValue(value_type, value)
        # 转换表头
        if dict_convert_head:
            for headName in dict_convert_head:
                convertHeadName = dict_convert_head[headName]
                if headName in rows_dict:
                    rows_dict[convertHeadName] = rows_dict.pop(headName)  # 字段名转换
                else:
                    rows_dict[convertHeadName] = ''  # 字段不存在,设置默认值
        # 过滤每行字段
        if filter_rows:
            isFilter = False
            try:
                args = (rows_dict,)
                if filter_rows_args:
                    args = args + filter_rows_args
                filterResult = filter_rows(*args)
                if filterResult is not None:
                    if type(filterResult) == bool:
                        isFilter = filterResult
                        if isFilter: rows_dict['filter_reason'] = ''
                    elif type(filterResult) == str:
                        isFilter = True
                        rows_dict['filter_reason'] = filterResult
                    elif type(filterResult) == tuple:
                        isFilter = filterResult[0]
                        rows_dict['filter_reason'] = filterResult[1]
            except Exception as e:
                import traceback
                traceback.print_exc()
                # print('过滤Excel rows,错误: %s' % e, rows_dict)
                isFilter = True
                rows_dict['filter_reason'] = str(e)
            if isFilter:
                filterContentList.append(rows_dict)
                # print('过滤数据',rows_dict)
                continue
        contentList.append(rows_dict)


def excelToBean(excelPath,dict_convert_head = None,filter_rows = None,filter_rows_args=None):
    """
     Excel列表转对象
    :param excelPath: excel全路径
    :return: list[{head:value,head2:value2...}]
    """
    import xlrd
    contentList = []
    filterContentList = []
    try:
        excel = xlrd.open_workbook(excelPath)
        for index in range(0,excel.nsheets):
            table = excel.sheets()[index]
            tableName = excel.sheet_names()[index]
            try:
                excelSheetRead(tableName,table,contentList,filterContentList,dict_convert_head,filter_rows,filter_rows_args)
            except:
                import traceback
                traceback.print_exc()
    except Exception as e:
        import traceback
        traceback.print_exc()
        print('读取excel文件,失败',excelPath,e)
    return contentList, filterContentList




def objToJson(obj, indent=None, customJSONEncoder=None):
    '''对象转JSON'''
    import json
    from lsptoolbox.common_class import DefaultJSONEncoder
    if not customJSONEncoder: customJSONEncoder = DefaultJSONEncoder
    jsonStr = json.dumps(obj, ensure_ascii=False, cls=customJSONEncoder, indent=indent)
    return jsonStr


def createResult(isSuccess = True, data = None , message = None, toJson = True, error=None):
    '''创建resp固定结构的返回值'''
    dict = {}
    if isSuccess:
        dict['code'] = 200
        dict['message'] = 'SUCCESS'
    else:
        dict['code'] = 0
        dict['message'] = 'FAIT'
    if message is not None:
        dict['message'] = message
    if data is not None:
        dict['erpdata'] = data
    if error is not None:
        dict['code'] = -200
        dict['message'] = 'ERROR'
        dict['error'] = str(error)
    if toJson:
        return objToJson(dict)
    return dict

def outToFileJsonFormat(dataBean,filePath):
    """json格式化保存对象到指定文件"""
    with open(filePath, 'w+',encoding='utf-8') as f:
        f.write(objToJson(dataBean,1))

def urlToLocalFile(url,localDirPath):
    '''下载url变成本地文件'''
    try:
        import requests,os
        if not os.path.exists(localDirPath):
            os.makedirs(localDirPath)
        resp = requests.get(url=url)
        fname = os.path.basename(url)
        fullPath = localDirPath + fname
        with open(fullPath, "wb+") as locFile:
            locFile.write(resp.content)
            return fullPath
    except Exception as e:
        print("URL转换本地文件",e)
    return None

def removeSymbol(sentence,remove_chars=None):
    '''过滤特殊符号'''
    if remove_chars is None:
        remove_chars = r'[·’!"\#$%&\'()（）＃！（）*+,-./:;<=>?\@，：?￥★、…．＞【】［］《》？“”‘’\[\\]^_`{|}~]+ⅡII'
    list = []
    for c in sentence.strip():
        list.append(c.strip(remove_chars))
    return ''.join(list)

def isEnglishWords(strs):
    import string
    for i in strs:
        if i not in string.ascii_lowercase + string.ascii_uppercase:
            return False
    return True

def delEnglishChar(str):
    '''移除英文字母'''
    nstr = ''
    for it in str:
        asciiCode = ord(it)
        if 65 <= asciiCode <= 90:continue
        if 97 <= asciiCode <= 122:continue
        if 0 <= asciiCode <= 9:continue
        nstr = nstr + it
    return nstr

def removeEnglishStr(str):
   # 判断是否全英文药品
   if isEnglishWords(str) is False:
       # 移除 英文字母干扰
       return delEnglishChar(str)
   return str

def recodeProcessPID(recode_pid_file):
    '''记录进程PID'''
    try:
        recode_pid = None
        import os
        if os.path.exists(recode_pid_file):
            # 存在已记录的PID文件
            with open(recode_pid_file, 'w') as f:
                try:
                    recode_pid = int(f.readline())
                except:
                    pass
        if recode_pid is not None:
            # 判断记录pid是否正在运行
            import psutil
            if psutil.pid_exists(recode_pid):
                return False
            # 记录当前程序的PID到文件
        with open(recode_pid_file, 'w') as f:
            f.write(str(os.getpid()))
            return True
    except Exception as e:
        print('尝试记录当前程序PID失败',recode_pid_file,e)
    return False

def writeFileContent(path,content):
    '''文件内容写入'''
    try:
        with open(path, 'w',encoding='utf-8') as f:
            f.write(str(content))
    except Exception as e:
        print('写入文件(%s)失败' % path, e)

def readFileContent(path,convertFunc=None):
    '''文件内容读取'''
    content = None
    try:
        import os
        if os.path.exists(path):
            with open(path, 'r',encoding='utf-8') as f:
                content = f.read()
            if convertFunc and content and len(content)>0:
                content = convertFunc(content)
    except Exception as e:
        print('读取文件(%s)失败' % path , e)
    return content

def calMillisecondsTimeDiff(access_start, access_end):
    '''计算毫秒数'''
    return int((access_end -access_start).seconds * 1000 + (access_end -access_start).microseconds / 1000)

def hommizationTimeDiff(access_start,access_end):
    '''格式化 时分秒'''
    delta = access_end - access_start
    seconds = int(round(delta.total_seconds()))
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    return "{:d}:{:02d}:{:02d}".format(hours, minutes, seconds)

def hommizationMillis(ts):
    millis = int(ts % 1000)
    seconds = int((ts/1000)%60)
    minutes = int((ts/(1000*60))%60)
    hours = int((ts/(1000*60*60))%24)
    day = int((ts/(1000*60*60*24))%360)
    info = ''
    if day>0: info+= str(day)+'天'
    if hours>0: info+= str(hours)+'小时'
    if minutes>0: info+= str(minutes)+'分钟'
    if seconds>0: info+= str(seconds)+'秒'
    if millis>0: info+= str(millis)+'毫秒'
    return  info

def jsonHandlePrev(obj):
    if isinstance(obj, dict):
        for k in list(obj):
            v = obj[k]
            if isinstance(k, bytes):
                obj[str(k, 'utf-8')] = obj.pop(k)
            jsonHandlePrev(v)
    if isinstance(obj, list):
        for it in obj:
            jsonHandlePrev(it)

def datetimeType(v):
    import datetime
    return datetime.datetime.strptime(v, "%Y-%m-%d %H:%M:%S") if type(v) == str else v

def stringSimilarityScore(str1, str2):
    import difflib, math
    score = difflib.SequenceMatcher(None, str(str1), str(str2)).quick_ratio()
    return math.ceil(score * 100)

def convertPercentage(float_val,percent=2):
    return str(format(float_val, '.'+str(percent)+'%'))

def decimalRound_old(value, digit):
    ''' float 四舍五入 '''
    from decimal import Decimal

    if value is None or str(value) == '':
        raise ValueError(' decimalRound value is '' or None ')

    result = str(Decimal(str(value)))

    if float(value) < 0:
        result = result[1:]
        if result != '':
            indexDec = result.find('.')
            if indexDec > 0:
                decimal = result[indexDec + 1:]
                decimalCount = len(decimal)
                if decimalCount > digit:
                    xiaoshu = result[indexDec + digit + 1]  # 第digit+1位小数
                    if int(xiaoshu) > 4:
                        result = str(float(value) * -1 + pow(10, digit * -1))
                        # 存在进位的可能，小数点会移位
                        indexDec = result.find('.')
                        result = result[:indexDec + digit + 1]
                    else:
                        result = result[:indexDec + digit + 1]
                else:
                    lens = digit - len(result[indexDec:]) + 1
                    for i in range(lens):
                        result += '0'
        result = float(result) * -1
    else:
        indexDec = result.find('.')
        if indexDec > 0:
            decimal = result[indexDec + 1:]
            decimalCount = len(decimal)
            if decimalCount > digit:
                xiaoshu = result[indexDec + digit + 1]  # 第digit+1位小数
                if int(xiaoshu) > 4:
                    result = str(float(value) + pow(10, digit * -1))
                    # 存在进位的可能，小数点会移位
                    indexDec = result.find('.')
                    result = result[:indexDec + digit + 1]
                else:
                    result = result[:indexDec + digit + 1]
            else:
                lens = digit - len(result[indexDec:]) + 1
                for i in range(lens):
                    result += '0'
    return float(result)

def decimalRound(value, digit=4):
    ''' dicimal 四舍五入 '''
    if value is None or str(value) == '':
        raise ValueError(' decimalRound value is '' or None ')
    import decimal
    from decimal import Decimal
    decimal.getcontext().rounding = "ROUND_HALF_UP"
    _digit =  '0.' + '0'*digit  if digit>0 else '0'
    return Decimal(str(value)).quantize(Decimal(_digit))

def roundStr(value,digit=4):
    format = '%.'+str(digit)+'f'
    return format % decimalRound(value,digit)

def div(a, b):
    ''' 除法'''
    try:
        from decimal import Decimal
        return float( Decimal(str(a)) / Decimal(str(b) ) )
    except:
        pass
    return 0

def mut(a, b):
    '''乘法'''
    try:
        from decimal import Decimal
        return float( Decimal(str(a)) * Decimal(str(b) ) )
    except:
        pass
    return 0

def add(a, b):
    '''加法'''
    try:
        from decimal import Decimal
        return float( Decimal(str(a)) + Decimal(str(b) ) )
    except:
        pass
    return 0

def sub(a, b):
    '''减法'''
    try:
        from decimal import Decimal
        return float( Decimal(str(a)) - Decimal(str(b) ) )
    except:
        pass
    return 0

def div_round(a, b, digit=4):
    return decimalRound(div(a, b), digit)

def mul_round(a, b, digit=4):
    return decimalRound(mut(a, b), digit)

def add_round(a, b, digit=4):
    return decimalRound(add(a, b), digit)

def sub_round(a, b, digit=4):
    return decimalRound(sub(a, b), digit)

def div_round_float(a, b, digit=4):
    return float(decimalRound(div(a, b), digit))

def mul_round_float(a, b, digit=4):
    return float(decimalRound(mut(a, b), digit))

def add_round_float(a, b, digit=4):
    return float(decimalRound(add(a, b), digit))

def sub_round_float(a, b, digit=4):
    return float(decimalRound(sub(a, b), digit))

def amt_divide(amt, num):
    '''计算金额及除不尽的余数'''
    if float(amt) <= 0 or int(num) <= 0: return 0.0, 0.0
    uprice = div_round(amt, num)
    indiv = sub(amt, mut(uprice, num))
    return uprice, indiv


def strConvertMD5(s):
    try:
        import hashlib
        myMd5 = hashlib.md5()
        myMd5.update(s.encode('utf-8'))
        myMd5_Digest = myMd5.hexdigest()
        return str(myMd5_Digest).upper()
    except: pass
    return s

def strToDatetime(datetimeStr):
    try:
        import datetime
        return datetime.datetime.strptime(datetimeStr, '%Y-%m-%d %H:%M:%S')
    except:
        return None

def strToDate(datetimeStr):
    try:
        import datetime
        return datetime.datetime.strptime(datetimeStr, '%Y-%m-%d')
    finally:
        return None

def join_string(*args):
    '''sql语句 字符串合并'''
    if len(args) <1: return None
    curStr = args[0]
    index = 1
    while(index<len(args)):
        curStr = args[index] % curStr
        index+=1
    return curStr


def getModuleFileSpecSuffixVars(module,suffix):
    '''获取指定模块文件的指定后缀的变量'''
    varList = []
    var_dict = dir(module)
    for varname in var_dict:
        if varname.endswith(suffix):
            var = getattr(module,varname)
            # print(' 获取varname ', varname, var)
            varList.append(var)
    return varList

def getMultipleModuleFileSpecSuffixVars(moduleList,suffix):
    '''获取多个模块文件的指定后缀的变量'''
    varList = []
    for module in moduleList:
        tempList = getModuleFileSpecSuffixVars(module,suffix)
        varList.extend(tempList)
    return varList


def socket_send_message(socket_conn,msg):
    '''socket 发送内容 处理粘包'''
    import struct
    bytes = msg.encode("utf-8")
    bytes_len = len(bytes)
    socket_conn.send(struct.pack('i', bytes_len))
    socket_conn.send(bytes)

def socket_receive_message(socket_conn):
    '''socket 接收内容 处理粘包'''
    import struct
    data = socket_conn.recv(4)
    len = struct.unpack('i', data)[0]
    bytes = socket_conn.recv(len)
    content = bytes.decode("utf-8")
    return content



'''
json.dumps 参数说明
skipkeys	是否跳过无法被JSON序列化的key（包括str, int, float, bool, None）	False
sort_keys	是否对数据按照key进行排序	False
ensure_ascii	输出保证将所有输入的非 ASCII 字符转义	True
allow_nan	是否允许JSON规范外的float数据(nan, inf, -inf)	True
default	是一个函数, 当某个value无法被序列化时, 对其调用该函数	None
indent	是一个正整数,  代表序列化后的缩进	None
separator	是一个格式为 (item_separator, key_separator) 的元组, 默认取值为 (', ', ': ')	None
check_circular	是否检查循环引用	True
'''

import json
import datetime
import decimal
class DefaultJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(obj, datetime.date):
            return obj.strftime("%Y-%m-%d")
        elif isinstance(obj, datetime.time):
            return obj.strftime("%H:%M:%S")
        elif isinstance(obj, datetime.timedelta):
            return (datetime.datetime.min + obj).time().strftime("%H:%M:%S")
        elif isinstance(obj, bytes):
            return obj.decode('utf-8')
        elif isinstance(obj, decimal.Decimal):
            return str(decimal.Decimal(obj).quantize(decimal.Decimal('0.0000')))
        elif isinstance(obj, list):
            return json.dumps(obj, cls=DefaultJSONEncoder, ensure_ascii=False)
        elif isinstance(obj, dict):
            return json.dumps(obj, cls=DefaultJSONEncoder, ensure_ascii=False)
        else:
            return json.JSONEncoder.default(self, obj)



# print(r'''
# .__                             __                .__ ___.
# |  |   ____________           _/  |_  ____   ____ |  |\_ |__   _______  ___
# |  |  /  ___/\____ \   ______ \   __\/  _ \ /  _ \|  | | __ \ /  _ \  \/  /
# |  |__\___ \ |  |_> > /_____/  |  | (  <_> |  <_> )  |_| \_\ (  <_> >    <
# |____/____  >|   __/           |__|  \____/ \____/|____/___  /\____/__/\_ \
#           \/ |__|                                          \/            \/
# ''')

import sys
from lsptoolbox.database_class_define import DatabaseClearThread, DatabaseManagerBean

print('thanks for using LSP tool box ,my phone is 15608447849 , sys.argv = ', sys.argv)

_db_mng_map = {}

def MYSQL_SET__GROUP_CONCAT_MAX_LEN(mysql_db, maxLen):
    '''设置mysql group_concat 函数 最大字符数量'''
    try:
        lines = mysql_db.query("show variables like 'group_concat_max_len';")
        if len(lines) == 1:
            curLen = int(lines[0][0])
            if curLen>=maxLen: return
        mysql_db.execute("SET GLOBAL group_concat_max_len = %d;" % maxLen)
        mysql_db.execute("SET SESSION group_concat_max_len = %d;" % maxLen)
        print('mysql concat_group max len: %d ' % maxLen)
    except: pass

def setting_database(database_obj):
    if database_obj.databaseType == 'mysql':
        MYSQL_SET__GROUP_CONCAT_MAX_LEN(database_obj, 4294967295)


def load_database_config(config):
    db_clear_thread = DatabaseClearThread()
    print('初始化数据库连接池空闲检测线程 ',db_clear_thread)
    for session in  config.sections():
        if session.startswith('dbload::'):
            mngName = session.replace('dbload::','',1)
            try:
                charset = config.get(session, 'charset')
            except:
                charset = 'utf8'

            databaseConnectBean = DatabaseManagerBean( db_clear_thread,
                                                       config.get(session, 'dbtype'),
                                                       config.get(session, 'host'),
                                                       config.get(session, 'port'),
                                                       config.get(session, 'username'),
                                                       config.get(session, 'password'),
                                                       config.get(session, 'database'),
                                                       charset
                                                       )
            setting_database(databaseConnectBean)
            _db_mng_map[mngName] = databaseConnectBean
            print("加载数据库 [ %s ],  %s" % (mngName, str(_db_mng_map[mngName])))

def getDatabaseOperation(databaseName):
    '''获取数据库'''
    mng = _db_mng_map.get(databaseName,None)
    if mng == None: raise ValueError("找不到指定数据库对象:%s" % databaseName)
    return mng

class CustomStdout:
    def __init__(self,appRootDir,cmd,recode=0):

        self.console = sys.stdout
        self.error = sys.stderr
        sys.stdout = self
        sys.stderr = self
        #sys.stderr = self
        self.appRootDir = appRootDir+'logs'
        # import datetime,os
        # self.procFlag = "%s_%s_%d" % (cmd,datetime.datetime.strftime(datetime.datetime.now(), '%H%M%S'),os.getpid())
        self.procFlag = cmd
        self.recode = recode
        self.msgText = []

    def write(self,outStr):
        try:
            if len(self.msgText) == 0:
                import datetime
                nowTimeStr = datetime.datetime.strftime(datetime.datetime.now(), '%Y-%m-%d %H:%M:%S')
                import os
                prevStr = "【%s】（%d）" % (nowTimeStr, os.getpid())
                self.msgText.append(prevStr)
                # self.console.write(prevStr)

            self.msgText.append(outStr)
            # self.console.write(outStr)
        except:
            import traceback
            traceback.print_exc()
        finally:
            try:
                if outStr.endswith('\n') and len(self.msgText)>0:
                    lineMsg = ''.join(self.msgText)
                    self.msgText.clear()
                    if 'NO CONSOLE' not in lineMsg:
                        self.console.write(lineMsg)
                    else:
                        lineMsg = lineMsg.replace('NO CONSOLE','*')
                    self.writeFile(lineMsg)
            except:
                import traceback
                traceback.print_exc()

    def flush(self): pass

    def writeFile(self, outStr):
        if self.recode == 0: return
        try:
            import datetime, os
            dir_path = self.appRootDir + '/' + datetime.datetime.strftime(datetime.datetime.now(), '%Y%m%d') +'/'
            if not os.path.exists(dir_path): os.makedirs(dir_path)
            file_path = dir_path +  self.procFlag + '.log'
            with open(file_path, 'a', encoding='utf-8') as f:
                f.write(outStr)
        except:
            import traceback
            traceback.print_exc()

# 加载配置
def load_config(project_name,cmd):
    import os
    import sys
    import configparser

    appRootDir = None
    if getattr(sys, 'frozen', False):
        appRootDir = os.path.dirname(sys.executable)
    elif __file__:
        appRootDir = os.path.dirname(os.path.realpath(sys.argv[0]))
    if appRootDir is None:
        print('程序运行路径异常,退出程序')
        os._exit(-1)

    if project_name in appRootDir:
        index = appRootDir.find(project_name)
        appRootDir = appRootDir[0:index + len(project_name)]
    if appRootDir.endswith('/') is False:
        appRootDir = appRootDir + '/'
    appRootDir = appRootDir.replace('\\', '/')
    printer = CustomStdout(appRootDir, cmd)
    print("应用(%s)(%d) 根目录路径: %s" % (project_name,os.getpid(), appRootDir))
    sys.path.append(appRootDir)

    configList = []
    for file in os.listdir(appRootDir):
        file_path = appRootDir + file
        if os.path.isfile(file_path) and os.path.splitext(file_path)[1] == '.ini':
            configList.append(file_path)


    config = configparser.RawConfigParser()
    config.read(filenames=configList, encoding='utf8')

    sections = config.sections()
    for section in sections:
        options = config.options(section)
        for option in options:
            var = config.get(section,option)
            if var.startswith('./'):
                var = appRootDir + var[2:]
                config.set(section,option,var)
            # print(section, option, var)

    printer.recode = config.getint('log','recode')

    load_database_config(config)

    return appRootDir, config



def windows_print_file(fp,printer):
    '''window系统文件打印'''
    try:
        import win32api
        import win32print
        # 设置默认打印机
        win32print.SetDefaultPrinter(printer)
        win32api.ShellExecute( 0,"print",fp,'/d:"%s"' % win32print.GetDefaultPrinter(),".", 0)
        return True
    except Exception as e:
        import traceback
        traceback.print_exc()
        return False

def saveAsDocx(file):
    '''保存文件未docx'''
    try:
        import win32com.client
        word = win32com.client.gencache.EnsureDispatch('Word.Application')
        word.Visible = False
        word.DisplayAlerts = False
        doc = word.Documents.Open(file, False)  # 打开文档，不提示转换确认框
        new_file = file.split('.')[0] + '.docx'
        doc.SaveAs(new_file, 12)
        doc.Close()
        return new_file
    except Exception as e:
        import traceback
        traceback.print_exc()
        return None

def combineAsDocx(master, sub):
    '''合并docx'''
    try:
        import os
        import docx
        import docxcompose.composer
        # 待合并文件必须存在
        if not os.path.exists(sub):
            return False

        # 主文件必须是docx格式,可以不存在
        if not master.endswith('.docx'):
            return False

        sub_docx = sub
        if not sub.endswith('.docx'):
            sub_docx = saveAsDocx(sub)

        if os.path.exists(master):
            doc_master = docx.Document(master)

            cp = docxcompose.composer.Composer(doc_master)
            cp.append(docx.Document(sub_docx))
        else:
            # master不存在，则sub直接给master
            doc_master = docx.Document(sub_docx)
        doc_master.save(master)
        return True
    except Exception as e:
        import traceback
        traceback.print_exc()
        return False